/* C implementation of drawBits for RetroBrowser to give speedup on jive */

#include "common.h"
#include "jive.h"

int jiveL_retro_drawbitsH(lua_State *L) {

	/* stack is:
		 1: new surface
		 2: old surface
		 3: new display bits - SB2 grfe bitmap string
		 4: old display bits - previous version
		 3: color 1
		 4: color 2
	*/

	Uint32 changed, word, mask, x, y, width, color1, color2, *new32, *old32, len, words;
	SDL_Surface *newsdl, *oldsdl;

	JiveSurface *newsrf = tolua_tousertype(L, 1, 0);
	JiveSurface *oldsrf = tolua_tousertype(L, 1, 0);
	new32  = (Uint32 *) luaL_checklstring(L, 3, &len);
	old32  = (Uint32 *) lua_tostring(L, 4);
	color1 = (Uint32) luaL_checkint(L, 5);
	color2 = (Uint32) luaL_checkint(L, 6);

	newsdl = newsrf->sdl;
	width  = newsdl->w;
	words  = len / 4;

	if (old32 == NULL) {
		/* draw full screen if no old bitmap to compare to */
		for (x = 0; x < words; ++x) {
			word = *new32;
			mask = 0x00000080;
			for (y = 0; y < 32; y++) {
				vlineColor(newsdl, x, 2*y, 2*y+1, word & mask ? color1 : color2);
				mask = mask & 0x01010101 ? mask << 15 : mask >> 1;
			}
			++new32;
		}
		for (x = words; x < width; ++x) {
			vlineColor(newsdl, x, 0, 32, color2);
		}			

	} else {
		/* copy old screen and only redraw changed pixels between old and new bits */
		oldsdl = oldsrf->sdl;
  
		SDL_BlitSurface(oldsdl, 0, newsdl, 0);

		for (x = 0; x < width; ++x) {
			if ((changed = *new32 ^ *old32)) {
				word = *new32;
				mask = 0x00000080;
				for (y = 0; y < 32; y++) {
					if (changed & mask) {
						vlineColor(newsdl, x, 2*y, 2*y+1, word & mask ? color1 : color2);
					}
					mask = mask & 0x01010101 ? mask << 15 : mask >> 1;
				}
			}
			++new32;
			++old32;
		}
	}

	return 0;
}


int jiveL_retro_drawbitsV(lua_State *L) {

	/* stack is:
		 1: new surface
		 2: old surface
		 3: new display bits - SB2 grfe bitmap string
		 4: old display bits - previous version
		 3: color 1
		 4: color 2
	*/

	Uint32 changed, word, mask, x, y, height, color1, color2, *new32, *old32, len, words;
	SDL_Surface *newsdl, *oldsdl;
	JiveSurface *newsrf = tolua_tousertype(L, 1, 0);
	JiveSurface *oldsrf = tolua_tousertype(L, 1, 0);
	new32  = (Uint32 *) luaL_checklstring(L, 3, &len);
	old32  = (Uint32 *) lua_tostring(L, 4);
	color1 = (Uint32) luaL_checkint(L, 5);
	color2 = (Uint32) luaL_checkint(L, 6);

	newsdl = newsrf->sdl;
	height = newsdl->h;
	words  = len / 4;

	if (old32 == NULL) {
		/* draw full screen if no old bitmap to compare to */
		for (y = 0; y < words; ++y) {
			word = *new32;
			mask = 0x01000000;
			for (x = 0; x < 32; x++) {
				hlineColor(newsdl, 2*x, 2*x+1, y, word & mask ? color1 : color2);
				mask = mask & 0x80808080 ? mask >> 15 : mask << 1;
			}
			++new32;
		}
		for (y = words; y < height; ++y) {
			hlineColor(newsdl, 0, 32, y, color2);
		}

	} else {
		/* copy old screen and only redraw changed pixels between old and new bits */
		oldsdl = oldsrf->sdl;
  
		SDL_BlitSurface(oldsdl, 0, newsdl, 0);

		for (y = 0; y < height; ++y) {
			if ((changed = *new32 ^ *old32)) {
				word = *new32;
				mask = 0x01000000;
				for (x = 0; x < 32; x++) {
					if (changed & mask) {
						hlineColor(newsdl, 2*x, 2*x+1, y, word & mask ? color1 : color2);
					}
					mask = mask & 0x80808080 ? mask >> 15 : mask << 1;
				}
			}
			++new32;
			++old32;
		}
	}

	return 0;
}


static const struct luaL_Reg retrolib [] = {
	{ "drawBitsH", jiveL_retro_drawbitsH },
	{ "drawBitsV", jiveL_retro_drawbitsV },
 	{ NULL, NULL }
};


int luaopen_applets_RetroBrowser_retro (lua_State *L) {
	luaL_register(L, "retro", retrolib);
	return 1;
}
